using System;
using System.Collections;
using System.Windows.Forms;

namespace RepPP.Toolkit.Window {
    //*********************************************************
	/// <summary>
	/// Connect the error reporter object to the Rep++ Field
	/// </summary>
    //*********************************************************
	public class FldErrorProvider : ErrorProvider
	{
	    private ArrayList           m_arrCtlInError;
	    private ErrorIconAlignment  m_eAlignment;
	    private int                 m_iIconPading;

        //*********************************************************
        //
		/// <summary>
		/// ctor:                       Class constructor
		/// </summary>
		/// <param name="eAlignment">   Error alignment</param>
		/// <param name="iIconPading">  Icon pading</param>
		//
        //*********************************************************
		public FldErrorProvider(ErrorIconAlignment eAlignment, int iIconPading) {
		    m_arrCtlInError = new ArrayList();
		    m_eAlignment    = eAlignment;
		    m_iIconPading   = iIconPading;
		}

        //*********************************************************
        //
		/// <summary>
		/// IconPading:             Distance between the icon and the field
		/// </summary>
		//
        //*********************************************************
		public int IconPading {
		    get {
		        return(m_iIconPading);
		    }
		    set {
		        m_iIconPading = value;
		    }
		}

        //*********************************************************
        //
        /// <summary>
        /// ErrorIconAlignment:     Alignment of the error icon
        /// </summary>
        //
        //*********************************************************
		public ErrorIconAlignment ErrorIconAlignment {
		    get {
		        return(m_eAlignment);
		    }
		    set {
		        m_eAlignment = value;
		    }
		}

        //*********************************************************
        //
        /// <summary>
        /// AddError:               Add an error text to a control
        /// </summary>
        /// <param name="ctl">      Control for which we add the error</param>
        /// <param name="strError"> Error text</param>
        //
        //*********************************************************
		public virtual void AddError(Control ctl, string strError) {
		    SetError(ctl, strError);
		    SetIconAlignment(ctl, m_eAlignment);
		    SetIconPadding(ctl, m_iIconPading);
		    m_arrCtlInError.Add(ctl);
		}

        //*********************************************************
        //
        /// <summary>
        /// AddError:                       Add an error text to a field control attached to the specified field
        /// </summary>
        /// <param name="ctlContainer">     Form or other container containing the field control</param>
        /// <param name="fld">              Field</param>
        /// <param name="strError">         Error text</param>
        //
        //*********************************************************
		public virtual void AddError(Control ctlContainer, RepPP.Field fld, string strError) {
		    foreach (Control ctl in ctlContainer.Controls) {
                if (IsControlAttachedToFieldB(ctl, fld)) {
		            AddError(ctl, strError);
		        }
		        AddError(ctl, fld, strError);
		    }
		}

        //*********************************************************
        //
        /// <summary>
        /// IsControlAttachedToFieldB:  Check if the control is a rep++ field or a toolit 
        ///                             rep++ field and is attached to the field
        /// </summary>
        /// <param name="ctl">          The control</param>
        /// <param name="fld">          The field</param>
        /// <returns>
        /// True:                       The control is attached to the field
        /// Flase:                      The control is not attached to the field
        /// </returns>
        //
        //*********************************************************
        protected virtual bool IsControlAttachedToFieldB(Control ctl, RepPP.Field fld)
        {
            bool bRetVal = false;

            if (ctl is RepPP.RepPPField) {
                bRetVal = ((ctl as RepPP.RepPPField).Field == fld);
            } else if (ctl is FldCheckBox) {
                bRetVal = ((ctl as FldCheckBox).Field == fld);
            } else if (ctl is FldComboBox) {
                bRetVal = ((ctl as FldComboBox).Field == fld);
            } else if (ctl is FldDateTimePicker) {
                bRetVal = ((ctl as FldDateTimePicker).Field == fld);
            } else if (ctl is FldListBox) {
                bRetVal = ((ctl as FldListBox).Field == fld);
            } else if (ctl is FldMonthCalendar) {
                bRetVal = ((ctl as FldMonthCalendar).Field == fld);
            } else if (ctl is FldTextBox) {
                bRetVal = ((ctl as FldTextBox).Field == fld);
            }
            return(bRetVal);
        }

        //*********************************************************
        //
        /// <summary>
        /// ResetError:                     Reset all the error shown using AddError
        /// </summary>
        //
        //*********************************************************
		public virtual void Reset() {
		    foreach (Control ctl in m_arrCtlInError) {
		        SetError(ctl, "");
		    }
		    m_arrCtlInError.Clear();
		}
	}
}
